/*
 * SplineCoaster.h
 *
 *  Created on: March 19, 2009
 *      Author: jima (referencing jfhamlin) 
 */

#ifndef SPLINECOASTER_H_
#define SPLINECOASTER_H_


#include <cmath>
#include <vector>

#include "algebra3.h"

using namespace std;


struct SplinePoint {
    vec3 point;
    double azimuth;

    SplinePoint() {}
    SplinePoint(vec3 pt, double az = 0) : point(pt), azimuth(az) {}

    static SplinePoint lerp(double t, SplinePoint &a, SplinePoint &b) {
        return SplinePoint( (1-t)*a.point + t*b.point, (1-t)*a.azimuth + t*b.azimuth );
    }

    static SplinePoint sampleBSpline(vector<SplinePoint*>& cps, double t, bool closed = true, int degree = 3);
};

class SplineCoaster {
public:
    SplineCoaster(string filename);
    SplineCoaster();

    // renders the coaster
    void Render(int samplesPerPt, double crossSectionScale=.2, int supportsPerPt=1, double supportSize=.1, double groundY=0.0);

    // helpful functions to get an up direction from a given forward direction
    bool getFrameUp(const vec3 &dir, const vec3 &canonicalUP, vec3 &up);
    vec3 getFrameUpWithoutFail(const vec3 &dir);
    void orientVectorInFrame(const vec3 &dir, double percent, double localAz, vec3 &inFrame);

    // sample the curve at a point
    SplinePoint sample(double t) {
        return SplinePoint::sampleBSpline(bsplinePts, t);
    }
    // get the forward direction
    vec3 sampleForward(double t) {
        return sample(t+.001).point - sample(t).point;
    }
    // get the up direction
    vec3 sampleUp(double t) {
        vec3 dir = sampleForward(t);
        vec3 up = getFrameUpWithoutFail(dir).normalize();
        orientVectorInFrame(-dir, fmod(t,1.0), sample(t).azimuth, up);
        return up;
    }

    // if there are no control points, that's bad
    bool bad() {
        if (bsplinePts.size() == 0)
            return true;
        return false;
    }

private:
    vector<SplinePoint*> bsplinePts; // control points for the bspline
    double globalAzimuth; // global azimuth rotates the whole frame
    double globalTwist; // twists the whole frame
    vector<vec2> profile; // the cross section to be swept

    // internal helper functions
    void sendVertex(vec3 &v);
    void createPolyline(vector<SplinePoint*> &pts, int totalSamples);
    void renderSupports(int supportsPerPt, double supportSize, double groundY);
    void renderSweep(vector<SplinePoint*> &pts, double crossSectionScale);
    void freePolyline(vector<SplinePoint*> &pts);
};


#endif

